/**
 * The MIT License (MIT)
 *
 * Copyright (C) 2013 Jaroslav Tulach <jaroslav.tulach@apidesign.org>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package com.oracle.chess.client.htmljava;

import com.oracle.chess.client.htmljava.Communication.AlertType;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import net.java.html.json.ComputedProperty;
import net.java.html.json.Function;
import net.java.html.json.Model;
import net.java.html.json.ModelOperation;
import net.java.html.json.OnPropertyChange;
import net.java.html.json.Property;
import net.java.html.sound.AudioClip;

@Model(className="Board", properties={
    @Property(name = "player", type = String.class),
    @Property(name = "gameId", type = String.class),
    @Property(name = "whitePlayer", type = String.class),
    @Property(name = "blackPlayer", type = String.class),
    @Property(name = "alert", type = Alert.class),
    @Property(name = "alertMessage", type = String.class),
    @Property(name = "status", type = String.class),
    @Property(name = "rows", type = Row.class, array = true),
    @Property(name = "turn", type = Color.class),
    @Property(name = "moves", type = Move.class, array = true),
    @Property(name = "pendingMove", type = Move.class),
    @Property(name = "active", type = boolean.class),
})
public class BoardModel {
    @ComputedProperty static String title(String status, String gameId) {
        String t = status != null ? status : gameId;
        if (t != null && t.length() > 10) {
            return t.substring(0, 10);
        }
        return t;
    }

    @ModelOperation static void updateSummary(Board b, String summary) {
        if (summary != null) {
            b.setStatus(summary);
        }
    }
    
    @ComputedProperty static boolean myTurn(String player, String whitePlayer, String blackPlayer, Color turn) {
        if (turn != null && player != null) switch (turn) {
            case B: return player.equals(blackPlayer);
            case W: return player.equals(whitePlayer);
        }
        return false;
    }
    
    @ComputedProperty static boolean justObserving(String player) {
        return player == null;
    }
    
    private static final AudioClip MOVE = AudioClip.create("sounds/move.mp3");
    @OnPropertyChange("moves") static void playMove() {
        MOVE.play();
    }

    private static final AudioClip CHECK = AudioClip.create("sounds/check.mp3");
    private static final AudioClip CHECKMATE = AudioClip.create("sounds/checkmate.mp3");
    @OnPropertyChange("alert") static void warnCheckAndMate(Board b) {
        if (b.getAlert() == null) {
            return;
        }
        if (b.getAlert().getType() == AlertType.CHECK) {
            CHECK.play();
        }
        if (b.getAlert().getType() == AlertType.CHECKMATE) {
            CHECKMATE.play();
        }
    }
    
    @Function static void selected(Board b, Square data) {
        if (!b.isMyTurn()) {
            b.setAlertMessage("Not your turn!");
            return;
        }
        
        Square previoslySelected = findSelectedSquare(b);
        if (previoslySelected == null) {
            if (data.getPiece() != null && data.getPieceColor() == b.getTurn()) {
                data.setSelected(true);
                Rules.computeAccessible(b, data);
            }
        } else {
            if (previoslySelected == data) {
                data.setSelected(false);
                Rules.computeAccessible(b, null);
                return;
            }
            if (data.getPiece() != null && data.getPieceColor() == previoslySelected.getPieceColor()) {
                previoslySelected.setSelected(false);
                data.setSelected(true);
                Rules.computeAccessible(b, data);
                return;
            }
            if (data.isAccessible()) {
                previoslySelected.setSelected(false);

                Move newMove = new Move();
                newMove.setFrom(previoslySelected.getPosition());
                newMove.setTo(data.getPosition());
                newMove.setRound(b.getMoves().size() / 2 + 1);
                newMove.setPiece(previoslySelected.getPiece());
                newMove.setTurn(previoslySelected.getPieceColor());
                newMove.setTakes(data.getPiece() != null);
                b.getMoves().add(newMove);
                b.setPendingMove(newMove);

                data.setPending(true);
                data.setPieceColor(previoslySelected.getPieceColor());
                data.setPiece(previoslySelected.getPiece());
                b.setTurn(null);
                previoslySelected.setPiece(null);
                previoslySelected.setPieceColor(null);
                Rules.computeAccessible(b, null);

                Request sm = new Request();
                sm.setMsg(MsgType.SendMove);
                sm.setGameId(b.getGameId());
                sm.setFrom(newMove.getFrom().getLocation());
                sm.setTo(newMove.getTo().getLocation());
                sm.setColor(newMove.getTurn());
                final UI ui = UIModel.findUI(b);
                if (ui != null) {
                    ui.sendMsg(sm);
                }
            }
        }
    }
    
    static class NextMove extends TimerTask {
        private static final Timer T = new Timer("Animate moves");
        private final Board b;
        private final Move m;

        public NextMove(Board b, Move m) {
            this.b = b;
            this.m = m;
            T.schedule(this, 1000);
        }
        
        
        @Override
        public void run() {
            b.showPosition(m);
        }
    }
    
    @ModelOperation @Function static void showPosition(Board b, Move data) {
        Rules.initBoard(b);
        boolean found = false;
        for (Move m : b.getMoves()) {
            if (found) {
                b.setTurn(null);
                new NextMove(b, m);
                return;
            }
            Square from = findSquare(b, (char)m.getFrom().getX(), m.getFrom().getY());
            Square to = findSquare(b, (char)m.getTo().getX(), m.getTo().getY());
            to.setPiece(from.getPiece());
            to.setPieceColor(from.getPieceColor());
            from.setPiece(null);
            from.setPieceColor(null);
            if (m == data) {
                found = true;
            }
        }
        b.setTurn(b.getMoves().size() % 2 == 0 ? Color.W : Color.B);
    }
    
    

    @Function static void rotateBoard(Board b) {
        Collections.reverse(b.getRows());
        for (Row r : b.getRows()) {
            Collections.reverse(r.getColumns());
        }
        Square sq = findSelectedSquare(b);
        if (sq != null) {
            sq.setSelected(false);
            Rules.computeAccessible(b, null);
        }
    }
    
    @ComputedProperty static boolean whiteTurn(Color turn) {
        return turn == Color.W;
    }

    @ComputedProperty static boolean blackTurn(Color turn) {
        return turn == Color.B;
    }
    
    @ComputedProperty static List<String> columnNames(List<Row> rows) {
        boolean whiteDown = rows.isEmpty() || rows.get(0).getY() == 8;
        String[] arr = new String[8];
        for (int i = 0; i < 8; i++) {
            String s;
            if (whiteDown) {
                s = "" + (char)('A' + i);
            } else {
                s = "" + (char)('H' - i);
            }
            arr[i] = s;
        }
        return Arrays.asList(arr);
    }
    
    static Square findSquare(Board b, char column, int row) {
        for (Row r : b.getRows()) {
            for (Square square : r.getColumns()) {
                if (square.getPosition().getX() == column && square.getPosition().getY() == row) {
                    return square;
                }
            }
        }
        return null;
    }

    private static Square findSquare(Board b, Position to) {
        return findSquare(b, (char)to.getX(), to.getY());
    }
    
    static Square findSelectedSquare(Board b) {
        for (Row row : b.getRows()) {
            for (Square square : row.getColumns()) {
                if (square.isSelected()) {
                    return square;
                }
            }
        }
        return null;
    }

    static void moveResponse(final Board b, final String errMsg, final List<String> whites, final List<String> blacks, final Color turn,
        Alert alert
    ) {
        if (errMsg != null) {
            b.getMoves().remove(b.getPendingMove());
            b.setPendingMove(null);
            b.setAlertMessage(errMsg);
        } else {
            b.setTurn(turn);
            b.setAlertMessage(null);
        }
        b.setAlert(alert);
        Rules.initBoard(b, whites, blacks, turn);
    }

    static void moveUpdate(final Board b, final Move move, final List<String> whites, final List<String> blacks, final Color turn, Alert alert) {
        final Square from = BoardModel.findSquare(b, move.getFrom());
        final Square to = BoardModel.findSquare(b, move.getTo());
        move.setPiece(from.getPiece());
        move.setTurn(from.getPieceColor());
        if (to.getPiece() != null) {
            move.setTakes(true);
        }
        b.setAlert(alert);
        b.setAlertMessage(alert == null ? null : alert.getMessage());
        move.setRound(b.getMoves().size() / 2 + 1);
        b.getMoves().add(move);
        Rules.initBoard(b, whites, blacks, turn);
    }

    @Model(className="Row", properties = {
        @Property(name = "columns", type = Square.class, array = true)
    })
    static class RowsImpl {
        @ComputedProperty static int y(List<Square> columns) {
            return columns.isEmpty() ? 0 : columns.get(0).getY();
        }
    }
    
    enum PieceType {
        PAWN(5), ROCK(2), KNIGHT(4), BISHOP(3), QUEEN(1), KING(0);

        final int entityIndex;
        
        PieceType(int ei) {
            this.entityIndex = ei;
        }

        static PieceType fromNotation(char notation) {
            switch (notation) {
                case 'R': return ROCK;
                case 'N': return KNIGHT;
                case 'B': return BISHOP;
                case 'Q': return QUEEN;
                case 'K': return KING;
                case 'P': return PAWN;
            }
            throw new IllegalStateException("Unexpected: " + notation);
        }
        
        String computeEntity(Color color) {
            if (color == null) {
                color = Color.W;
            }
            int base;
            switch (color) {
                case W: base = 12; break;
                case B: base = 18; break;
                default:
                    throw new AssertionError();
            }
            return "&#98" + String.valueOf(base + entityIndex) + ";";
        }
    }
    
    @Model(className="Position", properties = {
        @Property(name = "x", type = char.class),
        @Property(name = "y", type = int.class),
    })
    static class PositionImpl {
        @ComputedProperty static String location(int x, int y) {
            return "" + (char)(x - 'A' + 'a') + y;
        }
    }
    
    @Model(className="Square", properties = {
        @Property(name = "position", type = Position.class),
        @Property(name = "color", type = Color.class),
        @Property(name = "piece", type = PieceType.class),
        @Property(name = "pieceColor", type = Color.class),
        @Property(name = "selected", type = boolean.class),
        @Property(name = "accessible", type = boolean.class),
        @Property(name = "pending", type = boolean.class),
    })
    static class SquareModel {
        @ComputedProperty static String pieceEntity(
            PieceType piece, Color pieceColor
        ) {
            if (piece == null) {
                return "";
            }
            return piece.computeEntity(pieceColor);
        }
        
        @ComputedProperty static String squareColor(
            Color color, boolean selected, boolean accessible, boolean pending
        ) {
            if (selected) {
                return "selected";
            }
            if (accessible) {
                return "accessible";
            }
            if (pending) {
                return "pending";
            }
            
            if (color == null) {
                return "";
            } else {
                if (color == Color.W) {
                    return "white";
                } else {
                    return "black";
                }
            }
        }
        
        @ComputedProperty static char x(Position position) {
            return position == null ? 'A' : position.getX();
        }

        @ComputedProperty static int y(Position position) {
            return position == null ? 1 : position.getY();
        }
    }
    
    @Model(className = "Move", properties = {
        @Property(name = "round", type = int.class),
        @Property(name = "turn", type = Color.class),
        @Property(name = "piece", type = PieceType.class),
        @Property(name = "from", type = Position.class),
        @Property(name = "to", type = Position.class),
        @Property(name = "promoted", type = PieceType.class),
        @Property(name = "takes", type = boolean.class),
        @Property(name = "check", type = boolean.class),
    })
    static class MoveImpl {
        @ComputedProperty static boolean whiteMove(Color turn) {
            return turn == Color.W;
        }
        
        @ComputedProperty static String html(
            Position from, Position to, boolean takes, PieceType piece, Color turn
        ) {
            if (from == null || to == null) {
                return "";
            }
            StringBuilder sb = new StringBuilder();
            if (piece != null && piece != PieceType.PAWN) {
                sb.append(piece.computeEntity(turn));
            }
            
            sb.append(from.getLocation());
            if (takes) {
                sb.append("x");
            }
            sb.append(to.getLocation());
            return sb.toString();
        }

        static Move valueOf(String move) {
            move = move.toUpperCase();
            Move m = new Move();
            {
                Position p = new Position();
                p.setX(move.charAt(0));
                p.setY(move.charAt(1) - '0');
                m.setFrom(p);
            }
            {
                Position p = new Position();
                p.setX(move.charAt(2));
                p.setY(move.charAt(3) - '0');
                m.setTo(p);
            }
            return m;
        }
    }
}
